import argparse
import logging
import sys
from datetime import timedelta

import yaml

from email_forwarding_checker import __version__
from email_forwarding_checker.daemon import Daemon
from email_forwarding_checker.forwarding_checker import ForwardingChecker

__author__ = "verybadsoldier"
__copyright__ = "verybadsoldier"
__license__ = "GPLv3"

_logger = logging.getLogger(__name__)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def setdefault_recursively(tgt, defaults):
    for k in defaults:
        if isinstance(defaults[k], dict):  # if the current item is a dict,
            # expand it recursively
            setdefault_recursively(tgt.setdefault(k, {}), defaults[k])
        else:
            # ... otherwise simply set a default value if it's not set before
            tgt.setdefault(k, defaults[k])


def main(args):
    parser = argparse.ArgumentParser(description="Email Forwarding Checker")
    parser.add_argument(
        "--config_file", help="Config filename", type=str, default="config.yml"
    )

    parser.add_argument(
        "--daemon",
        "-d",
        action="store_true",
        help="Run as a daemon permanently reporting via MQTT",
    )

    parser.add_argument(
        "--emails",
        help="Overrides the emaiRun as a daemon permanently reporting via MQTT",
        type=str,
        nargs="*",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"email_forwarding_checker {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    config_defaults = dict(
        smtp=dict(port=587),
        mqtt=dict(host="localhost", port=1883, topic_base="email_forwarding_checker"),
        imap=dict(mailbox="inbox"),
        timeout=120,
        delete_emails=False,
        daemon=dict(check_interval_hours=24, run_now=False),
        email_timeout=120,
    )

    args = parser.parse_args(args)

    with open(args.config_file, "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)
        setdefault_recursively(config, config_defaults)

    if args.emails is not None and len(args.emails) > 0:
        config["emails"] = args.emails

    forwarding_checker = ForwardingChecker(
        smtp_sender=config["smtp"]["sender"],
        smtp_host=config["smtp"]["host"],
        smtp_port=config["smtp"]["port"],
        smtp_username=config["smtp"]["username"],
        smtp_password=config["smtp"]["password"],
        imap_host=config["imap"]["host"],
        imap_username=config["imap"]["username"],
        imap_password=config["imap"]["password"],
        imap_mailbox=config["imap"]["mailbox"],
        delete_emails=config["delete_emails"],
        email_timeout=config["email_timeout"],
        repeat_interval=config["repeat_interval"],
    )

    email_timeout = timedelta(seconds=config["email_timeout"])

    if args.daemon:
        setup_logging(logging.INFO if args.loglevel is None else args.loglevel)
        _logger.info("Starting in daemon with config: %s", str(config))
        d = Daemon(
            forwarding_checker,
            config["mqtt"]["host"],
            config["mqtt"]["port"],
            config["mqtt"]["topic_base"],
        )
        d.run(
            timedelta(minutes=config["daemon"]["check_interval_minutes"]),
            config["daemon"]["run_now"],
            config["emails"],
            email_timeout,
        )
    else:
        logging.getLogger().disabled = True
        report = forwarding_checker.check_multiple_emails(
            config["emails"], email_timeout
        )
        print(report)


def run():
    """Calls :func:`main` passing the CLI arguments extracted from :obj:`sys.argv`

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import logging

import paho.mqtt.client as mqtt

_logger = logging.getLogger(__name__)


class Mqtt:
    def __init__(self, hostname, port=1883):
        self._mqtthost = hostname
        self._mqttport = port

        self._client = mqtt.Client()

    def connect(self):
        _logger.info("Connecting to MQTT host '%s:%d'", self._mqtthost, self._mqttport)

        self._client.connect(self._mqtthost, self._mqttport, 60)
        self._client.loop_start()

    def disconnect(self):
        self._client.disconnect()

    def publish(self, topic: str, value: str):
        _logger.info("Publishing MQTT message to topic '%s' with value '%s'", topic, value)
        self._client.publish(topic, value)



import json
import logging
import time
from datetime import timedelta
from typing import List, Optional

import schedule

from email_forwarding_checker.forwarding_checker import ForwardingChecker
from email_forwarding_checker.mqtt import Mqtt

_logger = logging.getLogger(__name__)


class Daemon:
    def __init__(
        self,
        fc: ForwardingChecker,
        mqtt_host: str,
        mqtt_port: int,
        mqtt_topic_base: str,
    ) -> None:
        self._fc = fc
        self._mqtt = Mqtt(mqtt_host, mqtt_port)

        self._emails: Optional[List[str]] = None
        self._email_timeout: Optional[timedelta] = None
        self._mqtt_topic_base = mqtt_topic_base

    def _job(self):
        try:
            _logger.info("Starting mail check")
            assert self._emails is not None
            assert self._email_timeout is not None
            report = self._fc.check_multiple_emails(self._emails, self._email_timeout)

            self._mqtt.connect()

            json_object = json.dumps(report)

            _logger.info("Publishing check report to MQTT")
            self._mqtt.publish(self._mqtt_topic_base, json_object)
        except Exception:
            _logger.exception("Error in job execution")
        finally:
            _logger.info("Job execution finished")

    def run(
        self,
        run_interval: timedelta,
        run_now: bool,
        emails: List[str],
        email_timeout: timedelta,
    ):
        self._emails = emails
        self._email_timeout = email_timeout

        _logger.info("Scheduling job every %d...", run_interval.total_seconds())
        schedule.every(int(run_interval.total_seconds())).seconds.do(self._job)

        if run_now:
            _logger.info("Running job one time now")
            schedule.run_all()

        while True:
            schedule.run_pending()
            time.sleep(1)



import email
import imaplib
import logging
import smtplib
import time
from datetime import datetime, timedelta
from email.utils import parsedate_to_datetime
from typing import Dict, List

_logger = logging.getLogger(__name__)


class ForwardingChecker:
    def __init__(
        self,
        smtp_sender: str,
        smtp_host: str,
        smtp_port: int,
        smtp_username: str,
        smtp_password: str,
        imap_host: str,
        imap_username: str,
        imap_password: str,
        imap_mailbox: str,
        delete_emails: bool,
        email_timeout: int,
        repeat_interval: int,
    ) -> None:
        self._smtp_sender = smtp_sender
        self._smtp_username = smtp_username
        self._smtp_password = smtp_password
        self._smtp_host = smtp_host
        self._smtp_port = smtp_port
        self._imap_host = imap_host
        self._imap_username = imap_username
        self._imap_password = imap_password
        self._body = "This is an automated email to test if configured mail forwarding is working  - sent by email_forwarding_checker (https://github.com/verybadsoldier/email_forwarding_checker)"
        self._subject_base = "EMail Forward Test - email_forwarding_checker"
        self._delete_emails = delete_emails
        self._mailbox = imap_mailbox
        self._email_timeout = email_timeout
        self._repeat_interval = repeat_interval

    def check_multiple_emails(self, emails: List[str], email_timeout: timedelta) -> Dict[str, int]:
        report: Dict[str, int] = {}
        for addr in emails:
            num_retries = 1
            
            tokens = addr.split(":")
            addr = tokens[0]
            if len(tokens) > 1:
                num_retries = int(tokens[1])

            result = self.send_and_check_email(addr, email_timeout, num_retries)
            report[addr] = 1 if result else 0
        return report

    def send_and_check_email(self, dest_email: str, email_timeout: timedelta, num_retries: int = 1) -> bool:
        subject = f"{self._subject_base} - {dest_email}"

        for i in range(num_retries):
            _logger.info("%s - Starting try #%d/%d...", dest_email, i+1, num_retries)
            # Send the email
            start_time = datetime.now().astimezone()
            _logger.info("%s - Logging into SMTP server: %s", dest_email, self._smtp_host)
            with smtplib.SMTP(self._smtp_host, self._smtp_port) as server:
                server.starttls()
                server.login(self._smtp_username, self._smtp_password)

                body = self._body + f"\nSent on: {str(datetime.now().isoformat())}"

                message = f"Subject: {subject}\n\n{body}"
                _logger.info("%s - Sending email to %s", dest_email, dest_email)
                result = server.sendmail(self._smtp_sender, dest_email, message)
                if len(result) > 0:
                    _logger.error("Sending mail failed!")

            _logger.info("%s - Logging into IMAP server %s", dest_email ,self._imap_host)
            with imaplib.IMAP4_SSL(self._imap_host) as mail:
                mail.login(self._imap_username, self._imap_password)
                mail.select(self._mailbox)

                num_delete = 0
                while True:
                    now = datetime.now().astimezone()

                    diff = now - start_time
                    if diff > email_timeout:
                        _logger.info("%s - Timeout reached. Giving up", dest_email)
                        break

                    _logger.info("%s - Waiting for %d seconds for the mail to arrive... (Trying since %s/%s)", dest_email, self._repeat_interval, diff, email_timeout)
                    time.sleep(self._repeat_interval)

                    # Refresh inbox
                    mail.noop()

                    _logger.info("%s - Searching for the mail in the inbox...", dest_email)

                    #status, email_ids = mail.search(None, f'(SUBJECT "{subject}")')
                    status, email_ids = mail.uid('search', None, "UNSEEN")

                    if status != "OK":
                        raise RuntimeError("{dest_email} - Error IMAP search")

                    found = False
                    for email_uid in email_ids[0].split():
                        _, msg_data = mail.uid('fetch', email_uid, "(BODY.PEEK[HEADER.FIELDS (SUBJECT DATE)])")
                        if msg_data is None:
                            _logger.warning("%s - No email data found", dest_email)
                            continue
                        
                        response = msg_data[0]
                        if response is None:
                            _logger.warning("%s - No email response data found", dest_email)
                            continue

                        raw_email = response[1]
                        assert isinstance(raw_email, bytes)
                        msg = email.message_from_bytes(raw_email)

                        email_subject = msg["Subject"]
                        if email_subject != subject:
                            continue

                        _logger.debug("%s - Email with matching subject found", dest_email)

                        if self._delete_emails:
                            num_delete += 1
                            mail.uid('store', email_uid, "+FLAGS", "(\\Deleted)")

                        timestamp = parsedate_to_datetime(msg["Date"])
                        if timestamp >= start_time:
                            _logger.info("%s - Test mail found", dest_email)
                            found = True
                        else:
                            _logger.debug("%s - Found email too old", dest_email)


                    if num_delete > 0:
                        _logger.info("%s - Deleting marked emails", dest_email)
                        mail.expunge()

                    if found:
                        return True
                    else:
                        _logger.info("%s - Test mail not found", dest_email)
        _logger.info("%s - Test mail not found after all retries. Exiting...", dest_email)
        return False